from dotenv import load_dotenv
from environs import Env

env = Env()
load_dotenv()


class Config:
    CISCO_XLS_URL = env.str(
        "WLCL_CISCO_XLS_URL",
        "https://www.cisco.com/c/dam/assets/prod/wireless/wireless-compliance-tool/ComplianceStatus.xls"
    )
    SQLITE_URI = env.str("WLCL_SQLITE_URI", "wlcl.db")
    REFRESH_DAYS = env.int("WLCL_REFRESH_DAYS", 5)


__all__ = ["Config"]



from datetime import datetime

from .db import *
from .countries import get_country_iso2
from threading import Lock


class CiscoAPLookup:
    _refresh_lock = Lock()
    _refresh_date = None

    @classmethod
    def refresh_if_needed(cls):
        # class method to quickly check/perform updates as needed
        if not cls._refresh_date or cls._refresh_date < datetime.now():
            with cls._refresh_lock:
                if not cls._refresh_date or cls._refresh_date < datetime.now():
                    xls_data, cls._refresh_date = refresh_required()
                    if xls_data:
                        refresh_data(xls_data)

    @classmethod
    def models(cls) -> list[str]:
        cls.refresh_if_needed()
        return get_models()

    @classmethod
    def domain_for(cls, model: str, cn_iso2: str = None) -> list[str]:
        """
        Get list of possible regulatory domains for AP model and optionally country
        :param str model: AP Model name (without regulatory domain part)
        :param str cn_iso2: iso2 country - detects iso2 from full name, but with perf. impact.
        :rtype: list[str]
        """
        cls.refresh_if_needed()
        if cn_iso2 and len(cn_iso2) > 2:
            cn_iso2 = get_country_iso2(cn_iso2)
        return get_domain_for(model, cn_iso2)

    @classmethod
    def models_for(cls, model: str, cn_iso2: str = None) -> list[str]:
        """
        Get list of AP model PIDs with RD for model and optionally country
        :param str model: AP Model name (without regulatory domain part)
        :param str cn_iso2: iso2 country - detects iso2 from full name, but with perf. impact.
        :rtype: list[str]
        """
        cls.refresh_if_needed()
        if cn_iso2 and len(cn_iso2) > 2:
            cn_iso2 = get_country_iso2(cn_iso2)
        return get_models_for(model, cn_iso2)

    @classmethod
    def country_models(cls, model: str) -> list[str]:
        """
        Get list of unique AP model PIDs with RD for model
        :param str model: AP Model name (without regulatory domain part)
        :rtype: list[str]
        """
        cls.refresh_if_needed()
        return get_models_for(model)


__all__ = ["CiscoAPLookup"]



from re import compile, IGNORECASE
from io import BytesIO
import time
from openpyxl import load_workbook
from openpyxl.worksheet.worksheet import Worksheet
from openpyxl.workbook.workbook import Workbook
from ciscoaplookup.countries import (get_country_iso2)

platforms = {
    "Indoor": (2, 4),
    "Outdoor and Industrial": (2, 3),
}
model_pat = compile(r"^(AIR-|IW|C[9W]).*", IGNORECASE)
MAX_COL = 160
_cn_to_iso2_cache = {}


def get_book(xls_file: BytesIO) -> Workbook:
    book = load_workbook(xls_file, read_only=True, data_only=True)
    return book


def get_platform_sheet(wb: Workbook, platform: str) -> Worksheet:
    try:
        return wb[platform]
    except Exception as sheet_error:
        if "&" not in platform:
            raise sheet_error
        return wb[platform.replace("&", "and")]


def parse_row(raw_row: tuple[str], cols: dict[str, int]) -> dict[str, str]:
    vals = {}
    for header, col in cols.items():
        vals[header] = raw_row[col-1]
    return vals


def platform_indexes(platform_sheet: Worksheet) -> tuple[int, int]:
    global platforms
    if indexes := platforms.get(platform_sheet.title, None):
        rd_col = indexes[1]
        cn_col = indexes[0]
        return cn_col, rd_col
    else:
        raise Exception(f"Unable to locate sheet info for {platform_sheet.title}?")


def parse_models(platform_sheet: Worksheet, start_col: int) -> dict[str, list[int]]:
    models = {}
    try:
        for col in range(start_col, MAX_COL):
            txt = platform_sheet.cell(1, col).value
            if model_pat.search(txt):
                models[txt.upper()] = col
            else:
                break
    except IndexError:
        pass
    return models


def validate_headers(platform_sheet: Worksheet, cn_col: int, rd_col: int):
    h = {"Country": cn_col, "Regulatory Domain": rd_col}
    for header, col in h.items():
        if val := platform_sheet.cell(1, col).value:
            if val == header:
                continue  # all in order
        msg = f"Found unexpected header at 1,{col}: {val} vs {header}"
        raise ValueError(msg)


def parse_platform_models(wb: Workbook, platform: str) -> list[dict[str, object]]:
    global _cn_to_iso2_cache
    sheet = get_platform_sheet(wb, platform)
    cn_col, rd_col = platform_indexes(sheet)
    models = parse_models(sheet, rd_col+1)
    cols = models.copy()
    cols["CN"] = cn_col
    cols["RD"] = rd_col
    platform_models = []
    validate_headers(sheet, cn_col, rd_col)
    for raw_row in sheet.iter_rows(3+1, sheet.max_row+1, values_only=True):
        t1 = time.time()
        row = parse_row(raw_row, cols)
        cn = row.pop("CN")
        rd = row.pop("RD")
        if not cn or not rd:
            continue
        t2 = time.time()
        if cn_iso2 := _cn_to_iso2_cache.get(cn.lower(), None):
            pass
        else:
            cn_iso2 = get_country_iso2(cn)
            _cn_to_iso2_cache[cn.lower()] = cn_iso2
        t3 = time.time()
        pm_base = {"CN": cn_iso2, "RD": rd}
        for model, value in row.items():
            if value == "x":
                pm = pm_base.copy()
                pm["model"] = model
                platform_models.append(pm)
        # print(f"{cn}/{rd} in {time.time()-t1:.2f} s - row: {t2-t1:.2f}s, cn({cn}->{cn_iso2}): {t3-t2:.2f}s, row_scan: {time.time()-t3:.2f}s..")
    return platform_models


__all__ = ["parse_platform_models", "get_book", "platforms"]



import time
from base64 import b64encode
from datetime import datetime, timedelta
from hashlib import md5
from sqlite3 import connect, Connection, Date, OperationalError
from typing import Union

from ciscoaplookup.spreadsheet_parsing import get_book, platforms, parse_platform_models
from ciscoaplookup.config import Config


def init_db():
    c = _cnx()
    try:
        cur = c.execute("""SELECT name FROM sqlite_master WHERE type='table' AND name='file_hashes';""")
        result = cur.fetchone()
        if result is None:
            c.execute("""CREATE TABLE IF NOT EXISTS file_hashes (
                                               file VARCHAR(100) PRIMARY KEY NOT NULL,
                                               hash VARCHAR(200) NOT NULL,
                                               datetime int(4) NOT NULL);""")
            c.execute("""CREATE TABLE IF NOT EXISTS models (
                                                model VARCHAR(15) NOT NULL,
                                                rd VARCHAR(5) NOT NULL,
                                                cn VARCHAR(3) NOT NULL,
                                                PRIMARY KEY (model, rd, cn));
                                                """)
            c.commit()

    finally:
        if c:
            c.close()


def insert_models(models: list[dict[str, str]], prune: bool = False):
    stmt = "INSERT INTO models (model, rd, cn) VALUES (?, ?, ?)"
    row_values = [(row["model"], row["RD"], row["CN"]) for row in models]
    c = _cnx()
    try:
        if prune:
             c.execute("DELETE from models where 1=1;")
        c.executemany(stmt, row_values)
        c.commit()
    finally:
        if c:
            c.close()


def update_file_hash(filename: str, hash: str):
    c = _cnx()
    now = int(time.time())

    try:
        c.execute(f"INSERT INTO file_hashes ('file', 'hash', 'datetime') VALUES(?,?,?)", (filename, hash, now))
        c.commit()
    except Exception as e:
        c.execute(f"UPDATE file_hashes SET hash=?, datetime=? where file = '{filename}';", (hash, now))
        c.commit()
    finally:
        if c:
            c.close()


def get_file_hash(filename: str) -> tuple[str, datetime]:
    c = _cnx()
    try:
        if cur := c.execute(f"SELECT hash, datetime FROM file_hashes where file ='{filename}';"):
            row = next(cur)
            return row[0], row[1]
        raise StopIteration(f"Unable to find hash for {filename}")
    except OperationalError as e:
        raise StopIteration(e.args[0])
    finally:
        if c:
            c.close()


def get_models() -> list[str]:
    c = _cnx()
    try:
        cur = c.execute("""select model, count(*) from models group by model;""")
        return [r[0] for r in cur]
    finally:
        if c:
            c.close()


def get_domain_for(model: str, cn_iso2: str = None) -> list[str]:
    qry = f"SELECT rd, count(cn) FROM (SELECT rd, cn FROM models WHERE model = '{model}')"
    qry += " GROUP BY rd"
    if cn_iso2:
        qry += f",cn having cn = '{cn_iso2}'"
    c = _cnx()
    try:
        cur = c.execute(qry)
        return [r[0] for r in cur]
    finally:
        if c:
            c.close()


def get_models_for(model: str, country: str = None) -> list[str]:
    domains = get_domain_for(model, country)
    if not domains:
        msg = f"Unable to find any valid regulatory domains for {model} in {country}"
        raise ValueError(msg)
    return [f"{model.upper()}{dom}{'-K9' if model.upper().startswith('AIR-') else ''}" for dom in domains]


def get_country_models(model: str) -> list[str]:
    return get_models_for(model)


def _cnx() -> Connection:
    return connect(Config.SQLITE_URI)


def refresh_time_and_hash() -> tuple[datetime, str]:
    hash_ = None
    try:
        hash_, date = get_file_hash(Config.CISCO_XLS_URL)
        if date:
            date = datetime.fromtimestamp(date)
            return date + timedelta(days=Config.REFRESH_DAYS), hash_
    except StopIteration as e:
        pass

    return datetime.now() - timedelta(days=1), hash_  # refresh now


def refresh_required() -> tuple[Union[bool, bytearray], datetime]:
    refresh_date, hash_ = refresh_time_and_hash()
    import requests
    xls_data = requests.get(Config.CISCO_XLS_URL, allow_redirects=True).content
    if hash_ and hash_ == b64encode(md5(xls_data).digest()).decode("utf-8"):
        return False, refresh_date  # no change in file
    return xls_data, refresh_date


def refresh_data(xls_data: bytearray):
    from io import BytesIO
    new_md5 = b64encode(md5(xls_data).digest()).decode("utf-8")
    wb = get_book(BytesIO(xls_data))
    prune = True
    init_db()
    for platform in platforms:
        models = parse_platform_models(wb, platform)
        insert_models(models, prune)
        prune = False
    update_file_hash(Config.CISCO_XLS_URL, new_md5)


__all__ = ["get_models_for",
           "get_country_models",
           "get_domain_for",
           "get_file_hash",
           "update_file_hash",
           "insert_models",
           "init_db",
           "get_models",
           "refresh_time_and_hash",
           "refresh_data",
           "refresh_required"
           ]



from re import compile, IGNORECASE, Pattern
import country_converter as coco


def get_country_regex(country: str) -> Pattern:
    pat = coco.convert(names=country, to="regex", not_found=None)
    if not pat:
        raise ValueError("Unknown country: {}".format(country))

    return compile(pat, IGNORECASE)


def get_country_iso2(country: str) -> str:

    if cn := coco.convert(names=country, to="ISO2", not_found=None):
        return cn
    raise ValueError("Unknown country: {}".format(country))



from unittest import TestCase, main
import time
from ciscoaplookup import *
from config import Config
from db import refresh_data


class TestCiscoAPLookup(TestCase):
    def test_positive_models(self):
        cases = [
            ('CW9162I', 'Japan', '-Q'),
            ('C9115AXE', 'Belgium', '-E'),
            ('C9130AXI', 'Gibraltar', '-E'),
            ('AIR-AP1542I', 'Mexico', '-A-K9'),
            ('AIR-AP1562E', 'New Zealand', '-Z-K9'),
            ('IW3702-2E-X-K9', 'Barbados', '-N'),
            ('C9124AXI', 'Denmark', '-E'),
            ('C9136I', 'Denmark', '-E'),
            ('C9120AXI', 'Denmark', '-E'),
            ('C9120AXI', 'South Africa', '-E')

        ]
        for c in cases:
            t1 = time.time()
            res = CiscoAPLookup.models_for(c[0], c[1])
            exp = [c[0]+c[2]]
            self.assertEqual(exp, res, '{} doesn\'t match {} for {} / {}'.format(res, exp, c[0], c[1]))
            print(f"Tested {c[0]} in {c[1]}: {time.time()-t1:.2f}")

    def test_9120(self):
        models = []
        cfgs = []
        model_id = 223
        x = CiscoAPLookup.country_models('C9120AXE')
        for m in x:
            if m[-1:] in [ 'T', 'E']:
                continue
            models.append(f"({model_id}, '{m}', 'ap')")
            cfgs.append(f"('{m}', {model_id}, 1000, 'ap', '', '0,1,2,3,4,5,6', 1)")
            model_id += 1
        print(f"insert into tbl_model(`id`, `name`, `kind`) VALUES {', '.join(models)};")

        print(f"insert into tbl_device_configuration(`name`, `model_id`, `mbps`, `roles_csv`, `licenses_csv`, "
              f"`categories_csv`, `active`) VALUES {', '.join(cfgs)};")

    def test_fail_models(self):
        self.assertRaises(ValueError, CiscoAPLookup.models_for, 'gnyf', 'Denmark')  # invalid model
        self.assertRaises(ValueError, CiscoAPLookup.models_for, 'AIR-CAP1532I', 'Neverland')  # invalid country
        self.assertRaises(ValueError, CiscoAPLookup.models_for, 'AIR-CAP1552H', 'Vietnam')  # not possible

    def test_hest(self):
        self.assertTrue(len(CiscoAPLookup.models()) > 20)

    def test_countries(self):
        import pycountry
        lawless_countries = []

        for c in pycountry.countries:
            try:
                klaf = CiscoAPLookup.models_for('AIR-AP2802I', c.name)
            except ValueError as ve:
                print(f"{ve}: {c.name}")

    def test_repeated_update(self):
        import requests

        xls_data = requests.get(Config.CISCO_XLS_URL, allow_redirects=True).content
        refresh_data(xls_data)
        refresh_data(xls_data)


if __name__ == '__main__':
    main()
